/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef double ttype;

#define BIGTTYPE double
#define BIGTTYPE_MAX KDT_INFTY_DOUBLE
typedef double bigttype;

#define TTYPE_INTEGER 0

#define TTYPE_MIN -KDT_INFTY_DOUBLE
#define TTYPE_MAX  KDT_INFTY_DOUBLE
#define TTYPE_SQRT_MAX  1e154

#define TTYPE d
#define TTYPE_M d

