/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef u16 ttype;

#define BIGTTYPE u32
#define BIGTTYPE_MAX UINT32_MAX
typedef u32 bigttype;

#define TTYPE_INTEGER 1

#define TTYPE_MIN 0
#define TTYPE_MAX UINT16_MAX
#define TTYPE_SQRT_MAX UINT8_MAX

#define TTYPE s
#define TTYPE_M s
