/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef u32 etype;

#define ETYPE_INTEGER 1

#define ETYPE_MAX  0xffffffffu
#define ETYPE_MIN  0

#define ETYPE u
#define ETYPE_M u

//#define ETYPE_KDT_DATA  KDT_DATA_U32
