/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef double dtype;

#define DTYPE_INTEGER 0
#define DTYPE_DOUBLE  1

#define DTYPE_MAX  KDT_INFTY_DOUBLE
#define DTYPE_MIN -KDT_INFTY_DOUBLE

#define DTYPE d
#define DTYPE_M d

#define DTYPE_KDT_DATA  KDT_DATA_DOUBLE
