/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef u32 ttype;

#define BIGTTYPE u64
#define BIGTTYPE_MAX UINT64_MAX
typedef u64 bigttype;

#define TTYPE_INTEGER 1

#define TTYPE_MIN 0
#define TTYPE_MAX UINT32_MAX
#define TTYPE_SQRT_MAX UINT16_MAX

#define TTYPE u
#define TTYPE_M u
