/*
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

typedef float etype;

#define ETYPE_INTEGER 0

#define ETYPE_MAX  KDT_INFTY_FLOAT
#define ETYPE_MIN -KDT_INFTY_FLOAT

#define ETYPE f
#define ETYPE_M f

//#define ETYPE_KDT_DATA  KDT_DATA_FLOAT
